"""SIMP: A python programming environment for cellular and lattice gas automata.
http://simpstep.sourceforge.net

# Copyright (C) 2004 Ted Bach, BU Programmable Matter Group
# CVS Information 
# $Revision: 1.5.4.1 $
# $Source: /cvsroot/simpstep/simp/setup.py,v $




"""
# TODO
#  1) add self-test capabilities to setup.py

DOCLINES = __doc__.split("\n")

from distutils.core import setup
from distutils.command.install_data import install_data
from distutils.core import Extension

import sys, os, os.path
import glob

# ----------------------------------------------------------------
VERSION = "0.7.0"
#  Generate version info. Store it in simpinclude.py
f = open(os.path.join("Lib", "simpinclude.py"), "w")
f.write("""
# This file is generated by setup.py.  DO NOT EDIT.
version     = '%s'
""" % (VERSION))
f.close()
# =================================================================
CLASSIFIERS = """\
Development Status :: 0.6.11 - Beta
Intended Audience :: Science/Research
Intended Audience :: Developers
License :: OSI Approved 
Programming Language :: C
Programming Language :: Python
Topic :: Software Development 
Topic :: Scientific/Engineering
Operating System :: Microsoft :: Windows
Operating System :: POSIX
Operating System :: Unix
"""

import numpy # for numarray compilation paths

# the following works around a problem in recent versions of numarray that
# fail to sufficiently escape the backslash in include_dir and write
#    "c:\\python24\\Include\numarray" 
# rather than 
#    "c:\\python24\\Include\\numarray"
numarray_include_dir = numpy.get_numarray_include().replace("\n","\\n")
numpy_include_dir = numpy.get_include().replace("\n","\\n")


def main():
    if not hasattr(sys, 'version_info') or sys.version_info < (2,2,0,'final'):
        raise SystemExit, "Python 2.2.0 or later required to build simp."
    
    def getparams(**p): return p
    p = getparams(name ="simp",
          version = VERSION,
          description = DOCLINES[0],
          long_description = "\n".join(DOCLINES[2:]),
          url = "http://pm.bu.edu/step",
          download_url = "http://pm.bu.edu/step",
          author = "Ted Bach",
          author_email  = "tbach@bu.edu",
          maintainer = "Ted Bach",
          maintainer_email = "tbach@bu.edu",
          classifiers = filter(None, CLASSIFIERS.split("\n")),
          license = "http://pm.bu.edu",
          packages    = ["simp","simp.test","simp.stepmodules",
                         "simp.stepmodules.Pc",
                         "simp.stepmodules.PcThreaded",
                         "simp.stepmodules.PcCodeGen",
                         ],
          package_dir = {"simp":"Lib",
                "simp.test":"Lib/test",
                "simp.stepmodules":"Lib/stepmodules",
                "simp.stepmodules.Pc":"Lib/stepmodules/Pc",
                "simp.stepmodules.PcThreaded":"Lib/stepmodules/PcThreaded",
                "simp.stepmodules.PcCodeGen":"Lib/stepmodules/PcCodeGen",
                         },
          platforms   = ["Windows","Linux","Solaris","Mac OS-X"],
          ext_modules  = [Extension("simp.stepmodules.Pc._scan",
              ["Src/_scan.c"],
              include_dirs = [numpy_include_dir, numarray_include_dir, 'Src'],
                                    ),
                          ]
                  );
    if sys.version_info <= (2,3):
        del p["classifiers"]
        del p["download_url"]

    setup(**p)
    
if __name__ == "__main__":
    main()
